/*  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.

 *  You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.RandomAccessFile;


public class WadReader {
	private static class FileEntry {
		DirectoryEntry directory;
		String filename;
		short id;
		int unknown;
		int offset;
		int csize;
	}
	private static class DirectoryEntry {
		final DirectoryEntry parent;
		final String filename;
		int directoryCount;
		int fileCount;
		DirectoryEntry(DirectoryEntry parent, String filename) {
			this.parent = parent;
			this.filename = filename;
		}
		String getFullPath() {
			if(parent == null) return filename + "/";
			return parent.getFullPath() + filename + "/";
		}
	}
	public static void read(RandomAccessFile in, File out, ProgressDisplay display) throws IOException{
		int unknown1 = Util.read32(in);
		int unknown2 = Util.read32(in);
		display.update("Reading file names...", true, 0);
		int filenamesSize = Util.read32(in);
		byte[] filenamesTable = new byte[filenamesSize];
		in.read(filenamesTable);

		display.update("Reading file entries...", true, 0);
		int entryCount = Util.read32(in);
		FileEntry[] entries = new FileEntry[entryCount];
		for(int i = 0; i < entryCount; i++) {
			entries[i] = new FileEntry();
			entries[i].id = (short) i;
		}
		
		//skip over this useless data
		in.seek(in.getFilePointer() + entryCount * 8);

		display.update("Reading file offsets...", true, 0);
		int files, directories;
		files = Util.read32(in);
		directories = Util.read32(in);
		files = Util.read32(in);		
		for(int i = 0; i < files; i++) {
			entries[i].offset = Util.read32(in);
		}

		display.update("Reading directory structure...", true, 0);
		DirectoryEntry current = new DirectoryEntry(null, "");
		current.directoryCount = Integer.MAX_VALUE; current.fileCount = Integer.MAX_VALUE;
		for(int i = 0; i < files + directories; i++) {
			String filename = Util.readString(filenamesTable, Util.read32(in));
			int v1 = Util.read32(in);
			short v2 = Util.read16(in);
			in.readByte();
			boolean isFile = in.readByte() != 0;
			Util.read32(in);
			if(isFile) {
				FileEntry entry = entries[v2];
				while(current.fileCount == 0) current = current.parent;
				current.fileCount--;
				entry.directory = current;
				entry.filename = filename;
				entry.csize = v1;
			} else {
				while(current.directoryCount == 0) current = current.parent;
				current.directoryCount--;
				current = new DirectoryEntry(current, filename);
				current.fileCount = v1;
				current.directoryCount = v2;
			}
		}
		
		display.setMaximumProgress(entries.length);
		int counter = 0;
		for(FileEntry entry: entries) {
			OutputStream o = null;
			try {
				String filename = entry.directory.getFullPath() + entry.filename;
				display.update("Extracting " + filename + "...", false, counter);
				in.seek(entry.offset);
				byte[] buf = Util.readZlib(in, entry.csize);
				File f = new File(out.getAbsolutePath() + filename);
				f.getParentFile().mkdirs();
				o = new FileOutputStream(f);
				o.write(buf);
				o.close();
			} catch(Exception e) {
				System.err.println("Unable to extract file: " + entry.id);
				System.err.println("Total Entries: " + entries.length);
				e.printStackTrace();
			} finally {
				if(o != null) o.close();
			}
			counter++;
		}
		display.update("Finished.", true, entries.length);
		
	}
	
}
